# -*- coding: utf-8 -*-
import os
import sys
from mamonsu.lib.plugin import Plugin
from mamonsu.lib.queue import Queue


class LogSender(Plugin):

    Interval = 2
    _sender = True

    def __init__(self, config):
        super(LogSender, self).__init__(config)
        self.metric_log = config.fetch('metric_log', 'directory')
        if self.metric_log is None:
            self._enabled = False
        self._metric_log_fds = {}
        self.queue = Queue()
        self.max_queue_size = config.fetch('sender', 'queue', int)
        self.max_size = config.fetch('metric_log', 'max_size_mb', int)
        self.max_size = self.max_size * 1024 * 1024
        self._check_size_counter = 0

    def run(self, zbx):
        self._flush()

    def send(self, key, value, host=None, clock=None):
        metric = (key, value, host, clock)
        if self.queue.size() > self.max_queue_size:
            self.log.error('Queue size over limit, replace last metrics')
            self.queue.replace(metric)
        else:
            self.queue.add(metric)

    def _flush(self):
        metrics = self.queue.flush()
        if len(metrics) == 0:
            return
        for metric in metrics:
            self._write(metric)

    def _write(self, metric):

        key, value = metric[0], metric[1]
        host, clock = metric[2], metric[3]

        if not os.path.isdir(self.metric_log):
            try:
                os.makedirs(self.metric_log)
            except Exception as e:
                self.log.error('Create directory error: {0}'.format(e))
                sys.exit(7)

        if host is None:
            host = 'localhost'
        metric_log = os.path.join(
                self.metric_log, '{0}.log'.format(host))

        # rotate if reached max size limit
        if self._check_size_counter > 2:
            size = os.path.getsize(metric_log)
            if size > self.max_size:
                # rotate to file <METRIC_LOG>.archive
                backup_file = '{0}.archive'.format(metric_log)
                self.log.info(
                    'Move file {0} to {1} (max size limit reached: {2} b)'.
                    format(metric_log, backup_file, size))
                # close descriptor
                if host in self._metric_log_fds:
                    self._metric_log_fds[host].close()
                    del self._metric_log_fds[host]
                # rename
                os.rename(metric_log, backup_file)
            self._check_size_counter = 0
        self._check_size_counter += 1

        if host not in self._metric_log_fds:
            self._metric_log_fds[host] = open(metric_log, 'a')

        try:
            self._metric_log_fds[host].write("{0}\t{1}\t{2}\n".format(
                clock, value, key))
            self._metric_log_fds[host].flush()
        except Exception as e:
            self._metric_log_fds[host].close()
            del self._metric_log_fds[host]
            self.log.error('Write metric error: {0}'.format(e))



import mamonsu.lib.platform as platform

__import__('mamonsu.plugins.common')
__import__('mamonsu.plugins.pgsql')
if platform.LINUX:
    __import__('mamonsu.plugins.linux')
if platform.WINDOWS:
    __import__('mamonsu.plugins.windows')



import re
from mamonsu.lib.plugin import Plugin


class DiskStats(Plugin):

    # todo yaxis right 100%
    # bold line

    # Track only physical devices without logical partitions
    OnlyPhysicalDevices = True

    re_stat = re.compile('^(?:\s+\d+){2}\s+([\w\d]+) (.*)$')
    # rd_ios rd_merges rd_sectors rd_ticks
    # wr_ios wr_merges wr_sectors wr_ticks
    # ios_in_prog tot_ticks rq_ticks

    def run(self, zbx):
        with open('/proc/diskstats', 'r') as f:

            devices = []
            all_read, all_write = 0, 0

            for line in f:
                if re.search('(ram|loop)', line):
                    continue
                m = self.re_stat.match(line)
                if m is None:
                    continue
                dev, val = m.group(1), m.group(2)
                if self.OnlyPhysicalDevices and re.search('\d+$', dev):
                    continue
                val = [int(x) for x in val.split()]
                read, write, ticks = val[0], val[4], val[9]
                all_read += read
                all_write += write
                devices.append({'{#BLOCKDEVICE}': dev})

                zbx.send('system.disk.read[{0}]'.format(
                    dev), read, self.DELTA_SPEED)
                zbx.send('system.disk.write[{0}]'.format(
                    dev), write, self.DELTA_SPEED)
                zbx.send('system.disk.utilization[{0}]'.format(
                    dev), ticks/10, self.DELTA_SPEED)

            zbx.send('system.disk.all_read[]', all_read, self.DELTA_SPEED)
            zbx.send('system.disk.all_write[]', all_write, self.DELTA_SPEED)
            zbx.send('system.disk.discovery[]', zbx.json({'data': devices}))

    def items(self, template):
        return template.item({
            'name': 'Block devices: read requests',
            'key': 'system.disk.all_read[]'
        }) + template.item({
            'name': 'Block devices: write requests',
            'key': 'system.disk.all_write[]'
        })

    def graphs(self, template):
        items = [
            {'key': 'system.disk.all_read[]', 'color': 'CC0000'},
            {'key': 'system.disk.all_write[]', 'color': '0000CC'}
        ]
        graph = {
            'name': 'Block devices: read/write operations',
            'items': items}
        return template.graph(graph)

    def discovery_rules(self, template):

        rule = {
            'name': 'Block device discovery',
            'key': 'system.disk.discovery[]',
            'filter': '{#BLOCKDEVICE}:.*'
        }

        items = [
            {
                'key': 'system.disk.utilization[{#BLOCKDEVICE}]',
                'name': 'Block device {#BLOCKDEVICE}: utilization',
                'units': Plugin.UNITS.percent},
            {
                'key': 'system.disk.read[{#BLOCKDEVICE}]',
                'name': 'Block device {#BLOCKDEVICE}: read operations'},
            {
                'key': 'system.disk.write[{#BLOCKDEVICE}]',
                'name': 'Block device {#BLOCKDEVICE}: write operations'}]

        graphs = [{
            'name': 'Block device overview: {#BLOCKDEVICE}',
            'items': [{
                    'color': 'CC0000',
                    'key': 'system.disk.read[{#BLOCKDEVICE}]'},
                {
                    'color': '0000CC',
                    'key': 'system.disk.write[{#BLOCKDEVICE}]'},
                {
                    'yaxisside': 1,
                    'color': '00CC00',
                    'key': 'system.disk.utilization[{#BLOCKDEVICE}]'}]
        }]

        return template.discovery_rule(rule=rule, items=items, graphs=graphs)



from .core import (
    Warning, Bytea, DataError, DatabaseError, InterfaceError, ProgrammingError,
    Error, OperationalError, IntegrityError, InternalError, NotSupportedError,
    ArrayContentNotHomogenousError, ArrayContentEmptyError,
    ArrayDimensionsNotConsistentError, ArrayContentNotSupportedError, utc,
    Connection, Cursor, Binary, Date, DateFromTicks, Time, TimeFromTicks,
    Timestamp, TimestampFromTicks, BINARY, Interval)

# Copyright (c) 2007-2009, Mathieu Fenniak
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
# derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

__author__ = "Mathieu Fenniak"


def connect(
        user=None, host='localhost', unix_sock=None, port=5432, database=None,
        password=None, ssl=False, timeout=None,
        application_name=None, **kwargs):
    """Creates a connection to a PostgreSQL database.

    This function is part of the `DBAPI 2.0 specification
    <http://www.python.org/dev/peps/pep-0249/>`_; however, the arguments of the
    function are not defined by the specification.

    :param user:
        The username to connect to the PostgreSQL server with. If this is not
        provided, pg8000 looks first for the PGUSER then the USER environment
        variables.

        If your server character encoding is not ``ascii`` or ``utf8``, then
        you need to provide ``user`` as bytes, eg.
        ``"my_name".encode('EUC-JP')``.

    :keyword host:
        The hostname of the PostgreSQL server to connect with.  Providing this
        parameter is necessary for TCP/IP connections.  One of either ``host``
        or ``unix_sock`` must be provided. The default is ``localhost``.

    :keyword unix_sock:
        The path to the UNIX socket to access the database through, for
        example, ``'/tmp/.s.PGSQL.5432'``.  One of either ``host`` or
        ``unix_sock`` must be provided.

    :keyword port:
        The TCP/IP port of the PostgreSQL server instance.  This parameter
        defaults to ``5432``, the registered common port of PostgreSQL TCP/IP
        servers.

    :keyword database:
        The name of the database instance to connect with.  This parameter is
        optional; if omitted, the PostgreSQL server will assume the database
        name is the same as the username.

        If your server character encoding is not ``ascii`` or ``utf8``, then
        you need to provide ``database`` as bytes, eg.
        ``"my_db".encode('EUC-JP')``.

    :keyword password:
        The user password to connect to the server with.  This parameter is
        optional; if omitted and the database server requests password-based
        authentication, the connection will fail to open.  If this parameter
        is provided but not requested by the server, no error will occur.

    :keyword ssl:
        Use SSL encryption for TCP/IP sockets if ``True``.  Defaults to
        ``False``.

    :keyword timeout:
        Only used with Python 3, this is the time in seconds before the
        connection to the database will time out. The default is ``None`` which
        means no timeout.

    :rtype:
        A :class:`Connection` object.
    """
    return Connection(
        user, host, unix_sock, port, database,
        password, ssl, timeout, application_name)

apilevel = "2.0"
"""The DBAPI level supported, currently "2.0".

This property is part of the `DBAPI 2.0 specification
<http://www.python.org/dev/peps/pep-0249/>`_.
"""

threadsafety = 3
"""Integer constant stating the level of thread safety the DBAPI interface
supports.  This DBAPI module supports sharing the module, connections, and
cursors, resulting in a threadsafety value of 3.

This property is part of the `DBAPI 2.0 specification
<http://www.python.org/dev/peps/pep-0249/>`_.
"""

paramstyle = 'format'
"""String property stating the type of parameter marker formatting expected by
the interface.  This value defaults to "format", in which parameters are
marked in this format: "WHERE name=%s".

This property is part of the `DBAPI 2.0 specification
<http://www.python.org/dev/peps/pep-0249/>`_.

As an extension to the DBAPI specification, this value is not constant; it
can be changed to any of the following values:

    qmark
        Question mark style, eg. ``WHERE name=?``
    numeric
        Numeric positional style, eg. ``WHERE name=:1``
    named
        Named style, eg. ``WHERE name=:paramname``
    format
        printf format codes, eg. ``WHERE name=%s``
    pyformat
        Python format codes, eg. ``WHERE name=%(paramname)s``
"""

# I have no idea what this would be used for by a client app.  Should it be
# TEXT, VARCHAR, CHAR?  It will only compare against row_description's
# type_code if it is this one type.  It is the varchar type oid for now, this
# appears to match expectations in the DB API 2.0 compliance test suite.

STRING = 1043
"""String type oid."""


NUMBER = 1700
"""Numeric type oid"""

DATETIME = 1114
"""Timestamp type oid"""

ROWID = 26
"""ROWID type oid"""

__all__ = [
    Warning, Bytea, DataError, DatabaseError, connect, InterfaceError,
    ProgrammingError, Error, OperationalError, IntegrityError, InternalError,
    NotSupportedError, ArrayContentNotHomogenousError, ArrayContentEmptyError,
    ArrayDimensionsNotConsistentError, ArrayContentNotSupportedError, utc,
    Connection, Cursor, Binary, Date, DateFromTicks, Time, TimeFromTicks,
    Timestamp, TimestampFromTicks, BINARY, Interval]

"""Version string for pg8000.

    .. versionadded:: 1.9.11
"""



# -*- coding: utf-8 -*-

import subprocess
import sys
import time
import logging
import os
import re
import glob

from mamonsu import __version__

sudoWorking = None


def is_sudo_working():
    global sudoWorking
    if os.getuid() == 0:
        sudoWorking = True
        return
    if sudoWorking is None:
        sudoWorking = False
        p = subprocess.Popen(
            'sudo true',
            shell=True,
            stdin=None,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True)
        exec_time = 0
        while p.poll() is None:
            time.sleep(0.1)
            exec_time += 0.1
            if exec_time >= 1:
                return
        if p.returncode == 0:
            sudoWorking = True
    return sudoWorking


class Shell(object):

    # exit status of timeout code
    TimeoutCode = -1
    # local var
    _sudo_result = None

    def __init__(self, cmd, timeout=10, sudo=False):
        self.status = self.TimeoutCode
        self.cmd, self._real_command, self.sudo = cmd, cmd, sudo
        if sudo and is_sudo_working():
            self._real_command = 'sudo {0}'.format(cmd)
        self.stdout, self.stderr = '', ''
        self.wait_time, self.exec_time = timeout, 0
        p = subprocess.Popen(
            self._real_command,
            shell=True,
            stdin=None,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True)
        while p.poll() is None:
            time.sleep(0.1)
            self.exec_time += 0.1
            line = p.stdout.read()
            if not line == '':
                self.stdout += line.decode('utf-8')
            line = p.stderr.read()
            if not line == '':
                self.stderr += line.decode('utf-8')
            if self.wait_time > 0 and self.exec_time >= self.wait_time:
                return
        self.status = p.returncode
        for line in p.stdout.readlines():
            self.stdout += line.decode('utf-8')
        for line in p.stderr.readlines():
            self.stderr += line.decode('utf-8')
        self.stdout = self.stdout.strip()
        self.stderr = self.stderr.strip()

    def error(self):
        return 'Command `{0}` error code: {1}, sdterr: {2}'.format(
            self.cmd, self.status, self.stderr)


class SystemInfo(object):

    def __init__(self, args):
        self.args = args
        # disable sudo
        global sudoWorking
        if not self.args.use_sudo:
            sudoWorking = False
        logging.info('Test if sudo is working: {0}'.format(is_sudo_working()))
        logging.info('Collecting date...')
        self.date = self._fetch_date()
        logging.info('Collecting hostname...')
        self.hostname = self._fetch_hostname()
        logging.info('Collecting sysctl...')
        self.raw_sysctl = self._fetch_sysctl()
        logging.info('Collecting sysctl...')
        self.sysctl = self._parse_sysctl(self.raw_sysctl)
        logging.info('Collecting dmesg...')
        self.dmesg = self._fetch_dmesg()
        logging.info('Collecting lspci...')
        self.lspci = self._fetch_lspci()
        logging.info('Collecting release version...')
        self.system_release = self._fetch_release().strip()
        logging.info('Collecting kernel version...')
        self.kernel = self._fetch_kernel()
        logging.info('Collecting virtualization...')
        self.virtualization = self._fetch_virtualization(self.dmesg)
        logging.info('Collecting raid...')
        self.parsed_raid = self._parse_raid(self.lspci, self.dmesg)
        logging.info('Collecting uptime...')
        self.uptime = self._fetch_uptime()
        logging.info('Collecting cpu arch...')
        self.cpu_arch = self._fetch_cpu_arch()
        logging.info('Collecting os arch...')
        self.os_arch = self._fetch_os_arch()
        logging.info('Collecting dmidecode info...')
        self.dmidecode = self._fetch_dmidecode()
        logging.info('Collecting system info via dmidecode...')
        self.system_info = self._fetch_system_info()
        logging.info('Collecting mount points...')
        self.mount = self._fetch_mount()
        logging.info('Collecting df...')
        self.df = self._fetch_df()
        logging.info('Collecting cpu info...')
        self.cpu_info = self._fetch_cpu_info()
        logging.info('Parsing cpu info...')
        self.parsed_cpu_info = self._parse_cpu_info(self.cpu_info)
        logging.info('Collecting meminfo...')
        self.meminfo = self._fetch_meminfo()
        logging.info('Parsing meminfo...')
        self.parsed_meminfo = self._parse_meminfo(self.meminfo, self.sysctl)
        logging.info('Collecting disks info...')
        self.disks = self._fetch_disk_info()
        logging.info('Collecting iostat...')
        self.iostat = self._fetch_iostat()
        logging.info('Collecting lvs info...')
        self.lvs = self._fetch_lvs()
        logging.info('Collecting vgs info...')
        self.vgs = self._fetch_vgs()
        logging.info('Collecting mdstat info...')
        self.mdstat = self._fetch_mdstat()
        logging.info('Collecting systemd settings...')
        self.systemd = self._fetch_systemd()

    def printable_info(self):

        def format_header(info):
            return "\n###### {0:20} ###########################\n".format(info)

        def format_out(key, val):
            return "{0:40}|    {1}\n".format(key, val)

        out = ''
        out += format_header('Report')
        out += format_out('Version', __version__)
        out += format_out('Platform', sys.platform)
        out += format_out('Python', sys.version_info)
        out += format_header('System')
        out += format_out('Date', self.date)
        out += format_out('Host', self.hostname)
        out += format_out('Uptime', self.uptime)
        out += format_out('System', self.system_info['TOTAL'])
        out += format_out('Serial', self.system_info['SERIAL'])
        out += format_out('Release', self.system_release)
        out += format_out('Kernel', self.kernel)
        out += format_out('Arch', 'CPU = {0}, OS = {1}'.format(
            self.cpu_arch, self.os_arch))
        out += format_out('Virt', self.virtualization)
        out += format_header('Processors')
        out += format_out('Total', self.parsed_cpu_info['_TOTAL'])
        out += format_out('Speed', self.parsed_cpu_info['speed'])
        out += format_out('Model', self.parsed_cpu_info['model'])
        out += format_out('Cache', self.parsed_cpu_info['cache'])
        out += format_header('Memory')
        out += format_out('Total', self.parsed_meminfo['_TOTAL'])
        out += format_out('Cached', self.parsed_meminfo['_CACHED'])
        out += format_out('Dirty', self.parsed_meminfo['_DIRTY'])
        out += format_out('Dirty ratio', self.parsed_meminfo['_DIRTY_RATIO'])
        out += format_out('Dirty bytes', self.parsed_meminfo['_DIRTY_BYTES'])
        out += format_out('Swap', self.parsed_meminfo['_SWAP'])
        out += format_out('Swappines', self.parsed_meminfo['_SWAPPINESS'])
        out += format_header('System settings')
        for k in self.systemd['_main']:
            out += format_out(k, self.systemd['_main'][k])
        out += format_header('Mount')
        out += self.df + "\n"
        out += format_header('Disks')
        for disk in self.disks:
            out += format_out(disk, 'Scheduler: {0} Queue: {1}'.format(
                self.disks[disk]['scheduler'],
                self.disks[disk]['nr_requests']))
        out += format_header('IOstat')
        out += self.iostat + "\n"
        out += format_header('LVM')
        out += self.vgs + "\n"
        out += self.lvs + "\n"
        out += format_header('Raid')
        for raid in self.parsed_raid:
            out += format_out('Controller', raid)
        return out

    def store_raw(self):
        def format_out(info, val):
            return "# {0} ##################################\n{1}\n".format(
                info, val)
        out = format_out('SYSCTL', self.raw_sysctl)
        out += format_out('DMESG', self.dmesg)
        out += format_out('LSPCI', self.lspci)
        out += format_out('CPUINFO', self.cpu_info)
        out += format_out('MEMINFO', self.meminfo)
        out += format_out('DMIDECODE', self.dmidecode)
        out += format_out('DF', self.df)
        out += format_out('MOUNT', self.mount)
        out += format_out('MDSTAT', self.mdstat)
        out += format_out('IOSTAT', self.iostat)
        out += format_out('LVS', self.lvs)
        out += format_out('VGS', self.vgs)
        return out

    def collect(self):
        info = self.printable_info()
        logging.error("\n{0}\n".format(self.store_raw()))
        return info

    _suffixes = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']

    def _humansize(self, nbytes):
        if nbytes == 0:
            return '0 B'
        i = 0
        while nbytes >= 1024 and i < len(self._suffixes)-1:
            nbytes /= 1024.
            i += 1
        f = ('%.2f' % nbytes).rstrip('0').rstrip('.')
        return '%s %s' % (f, self._suffixes[i])

    def _remove_duplicates(self, values):
        output = []
        seen = set()
        for value in values:
            if value not in seen:
                output.append(value)
                seen.add(value)
        return output

    def _fetch_hostname(self):
        shell = Shell('uname -n')
        if shell.status == 0:
            return shell.stdout
        else:
            logging.error(shell.error())
        return 'N/A'

    def _fetch_date(self):
        result = ''
        shell = Shell("date -u +'%F %T UTC'")
        if shell.status == 0:
            result += shell.stdout
        shell = Shell("date +'%Z %z'")
        if shell.status == 0:
            result = "{0} (local TZ:{1})".format(result, shell.stdout)
        return result

    def _fetch_sysctl(self):
        shell = Shell('sysctl -a')
        if shell.status == 0:
            return shell.stdout
        else:
            logging.error(shell.error())
        return ''

    def _parse_sysctl(self, sysctl):
        result = {}
        if sysctl == '':
            return result
        for out in sysctl.split("\n"):
            try:
                k, v = out.split(" = ")
                result[k] = v
            except:
                logging.error(
                    "unexpected `systcl -a` output: '{0}'".format(out))
                continue
        return result

    def _fetch_dmesg(self):
        try:
            if os.path.isfile('/var/log/dmesg'):
                with open('/var/log/dmesg', 'r') as content_file:
                    content = content_file.read()
                    return content
        except:
            pass
        shell = Shell('dmesg')
        if shell.status == 0:
            return shell.stdout
        else:
            logging.error(shell.error())
            return ''

    def _fetch_virtualization(self, dmesg):
        if dmesg == '':
            return 'N/A'
        if os.path.isfile('/proc/user_beancounters'):
            return 'OpenVZ/Virtuozzo'
        if re.search(r'vmware', dmesg, re.I):
            return 'VMWare'
        if re.search(r'vmxnet', dmesg, re.I):
            return 'VMWare'
        if re.search(r'paravirtualized kernel on vmi', dmesg, re.I):
            return 'VMWare'
        if re.search(r'Xen virtual console', dmesg, re.I):
            return 'Xen'
        if re.search(r'paravirtualized kernel on xen', dmesg, re.I):
            return 'Xen'
        if re.search(r'qemu', dmesg, re.I):
            return 'QEmu'
        if re.search(r'paravirtualized kernel on KVM', dmesg, re.I):
            return 'KVM'
        if re.search(r'VBOX', dmesg, re.I):
            return 'VirtualBox'
        if re.search(r'hd.: Virtual .., ATA.*drive', dmesg, re.I):
            return 'Microsoft VirtualPC'
        return 'N/A'

    def _fetch_lspci(self):
        shell = Shell('lspci')
        if shell.status == 0:
            return shell.stdout
        else:
            logging.error(shell.error())
            return ''

    def _fetch_release(self):
        for file in ['/etc/fedora-release', '/etc/redhat-release',
                     '/etc/system-release']:
                if os.path.isfile(file):
                    try:
                        with open(file, 'r') as content_file:
                            content = content_file.read()
                            return content
                    except:
                        pass
        if os.path.isfile('/etc/lsb-release'):
            try:
                with open('/etc/lsb-release', 'r') as f:
                    for line in f:
                        if not re.search(r'DISTRIB_DESCRIPTION', line):
                            continue
                        _, content = line.split('=')
                        return content
            except:
                pass
        if os.path.isfile('/etc/debian_version'):
            try:
                with open('/etc/debian_version', 'r') as f:
                    content = content_file.read()
                    return 'Debian-based version {0}'.format(content)
            except:
                pass
        return 'Unknown'

    def _fetch_kernel(self):
        shell = Shell('uname -r')
        if shell.status == 0:
            return shell.stdout
        else:
            logging.error(shell.error())
            return ''

    def _parse_raid(self, lspci, dmesg):
        controllers = []
        if lspci != '':
            if re.search(
                r'RAID bus controller: LSI Logic / Symbios Logic MegaRAID SAS',
                    lspci, re.I):
                controllers.append('LSI Logic MegaRAID SAS')
            if re.search(
                r'RAID bus controller: LSI Logic / Symbios Logic LSI MegaSAS',
                    lspci, re.I):
                controllers.append('LSI Logic MegaRAID SAS')
            if re.search(
                r'Fusion-MPT SAS',
                    lspci, re.I):
                controllers.append('Fusion-MPT SAS')
            if re.search(
                r'RAID bus controller: LSI Logic / Symbios Logic Unknown',
                    lspci, re.I):
                controllers.append('LSI Logic Unknown')
            if re.search(
                r'RAID bus controller: Adaptec AAC-RAID',
                    lspci, re.I):
                controllers.append('AACRAID')
            if re.search(
                r'3ware [0-9]* Storage Controller',
                    lspci, re.I):
                controllers.append('3Ware')
            if re.search(
                r'Hewlett-Packard Company Smart Array',
                    lspci, re.I):
                controllers.append('HP Smart Array')
            if re.search(
                r'Hewlett-Packard Company Smart Array',
                    lspci, re.I):
                controllers.append('HP Smart Array')
        if dmesg != '':
            if re.search(r'scsi[0-9].*: .*megaraid', dmesg, re.I):
                controllers.append('LSI Logic MegaRAID SAS')
            if re.search(r'Fusion MPT SAS', dmesg):
                controllers.append('Fusion-MPT SAS')
            if re.search(r'scsi[0-9].*: .*aacraid', dmesg, re.I):
                controllers.append('AACRAID')
            if re.search(
                r'scsi[0-9].*: .*3ware [0-9]* Storage Controller',
                    dmesg, re.I):
                controllers.append('3Ware')
        return self._remove_duplicates(controllers)

    def _fetch_uptime(self):
        shell = Shell('uptime')
        if not shell.status == 0:
            logging.error(shell.error())
        return shell.stdout

    def _fetch_cpu_arch(self):
        if os.path.isfile('/proc/cpuinfo'):
            try:
                with open('/proc/cpuinfo', 'r') as content_file:
                    content = content_file.read()
                    if re.search(r' lm ', content):
                        return '64-bit'
                    else:
                        return '32-bit'
            except:
                pass
        return 'N/A'

    def _fetch_os_arch(self):
        shell = Shell('getconf LONG_BIT')
        if shell.status == 0:
            if re.search('64', shell.stdout):
                return '64-bit'
            if re.search('32', shell.stdout):
                return '32-bit'
        else:
            logging.error(shell.error())
        return 'N/A'

    def _fetch_dmidecode(self):
        shell = Shell('dmidecode', sudo=True)
        if shell.status == 0:
            return shell.stdout
        else:
            logging.error(shell.error())
            return ''

    def _fetch_system_info(self):

        def dmidecode(key):
            shell = Shell('dmidecode -s "{0}"'.format(key), sudo=True)
            if shell.status == 0:
                return shell.stdout
            else:
                logging.error(shell.error())
                return ''

        result = {}
        result['vendor'] = dmidecode('system-manufacturer')
        result['product'] = dmidecode('system-product-name')
        result['version'] = dmidecode('system-version')
        result['chassis'] = dmidecode('chassis-type')
        result['SERIAL'] = dmidecode('system-serial-number')
        result['TOTAL'] = '{0}; {1}; {2} ({3})'.format(
            result['vendor'], result['product'],
            result['chassis'], result['version'])
        return result

    def _fetch_mount(self):
        shell = Shell('mount')
        if shell.status == 0:
            return shell.stdout
        else:
            logging.error(shell.error())
            return ''

    def _fetch_df(self):
        shell = Shell('df -h -P')
        if shell.status == 0:
            return shell.stdout
        else:
            logging.error(shell.error())
            return ''

    def _fetch_cpu_info(self):
        if os.path.isfile('/proc/cpuinfo'):
            try:
                with open('/proc/cpuinfo', 'r') as content_file:
                    content = content_file.read()
                    return content
            except:
                logging.error('Can\'t read /proc/cpuinfo')
                return ''
        else:
            logging.error('Can\'t find /proc/cpuinfo')
        return ''

    def _parse_cpu_info(self, info):

        def fetch_first(reg, info):
            val = re.search(reg, info, re.M)
            if val is not None:
                return val.group(1)
            else:
                return 'N/A'

        if info == '':
            return {}

        result = {}
        result['virtual'] = len(
            re.findall(r'(^|\n)processor', info))
        result['physical'] = len(self._remove_duplicates(
            re.findall(
                r'^physical id\s+\:\s+(\d+)', info, re.M)))
        cores = re.search(
            r'^cpu cores\s+\:\s+(\d+)', info, re.M)
        if cores is not None:
            result['cores'] = int(cores.group(1))
        else:
            result['cores'] = 0
        if result['physical'] == 0:
            result['physical'] = result['virtual']
        result['hyperthreading'] = False
        if result['cores'] > 0:
            if result['cores'] < result['virtual']:
                result['hyperthreading'] = True
        result['model'] = fetch_first(r'model name\s+\:\s+(.*)$', info)
        result['cache'] = fetch_first(r'cache size\s+\:\s+(.*)$', info)
        result['speed'] = fetch_first(
            r'^cpu MHz\s+\:\s+(\d+\.\d+)$', info) + ' MHz'
        result['_TOTAL'] = 'physical = {0}, cores = {1}, '\
            'virtual = {2}, hyperthreading = {3}'.format(
                result['physical'], result['cores'],
                result['virtual'], result['hyperthreading']
            )
        return result

    def _fetch_meminfo(self):
        if os.path.isfile('/proc/meminfo'):
            try:
                with open('/proc/meminfo', 'r') as content_file:
                    content = content_file.read()
                    return content
            except:
                logging.error('Can\'t read /proc/meminfo')
                return ''
        else:
            logging.error('Can\'t find /proc/meminfo')
        return ''

    def _parse_meminfo(self, data, sysctl):
        if data == '':
            return {}
        result = {}
        for info in re.findall(r'^(\S+)\:\s+(\d+)\s+kB$', data, re.M):
            result[info[0]] = int(info[1])*1024
        result['_TOTAL'] = ''
        if 'MemTotal' in result:
            result['_TOTAL'] = self._humansize(result['MemTotal'])
        result['_SWAP'] = ''
        if 'SwapTotal' in result:
            result['_SWAP'] = self._humansize(result['SwapTotal'])
        result['_CACHED'] = ''
        if 'Cached' in result:
            result['_CACHED'] = self._humansize(result['Cached'])
        result['_DIRTY'] = ''
        if 'Dirty' in result:
            result['_DIRTY'] = self._humansize(result['Dirty'])
        result['_SWAPPINESS'] = ''
        if 'vm.swappiness' in sysctl:
            result['_SWAPPINESS'] = sysctl['vm.swappiness']
        result['_DIRTY_RATIO'] = '0 0'
        if 'vm.dirty_ratio' in sysctl:
            if 'vm.dirty_background_ratio' in sysctl:
                result['_DIRTY_RATIO'] = '{0} {1}'.format(
                    sysctl['vm.dirty_ratio'],
                    sysctl['vm.dirty_background_ratio'])
        result['_DIRTY_BYTES'] = '0 0'
        if 'vm.dirty_bytes' in sysctl:
            if 'vm.dirty_background_bytes' in sysctl:
                result['_DIRTY_BYTES'] = '{0} {1}'.format(
                    sysctl['vm.dirty_bytes'],
                    sysctl['vm.dirty_background_bytes'])
        return result

    def _fetch_disk_info(self):
        result = {}
        for block in glob.glob('/sys/block/*'):
            data, disk = {}, block.replace('/sys/block/', '')
            if re.search('(ram|loop)', disk):
                continue
            scheduler = '{0}/queue/scheduler'.format(block)
            nr_requests = '{0}/queue/nr_requests'.format(block)
            try:
                if os.path.isfile(scheduler):
                    data['scheduler'] = open(scheduler, 'r').read().strip()
                if os.path.isfile(nr_requests):
                    data['nr_requests'] = open(nr_requests, 'r').read().strip()
                shell = Shell('fdisk -l /dev/{0}'.format(disk), sudo=True)
                if shell.status == 0:
                    data['fdisk'] = shell.stdout
                else:
                    logging.error(shell.error())
                    data['fdisk'] = ''
            except:
                continue
            result[disk] = data
        return result

    def _fetch_iostat(self):
        shell = Shell('iostat -x -N -m 1 2', timeout=3)
        if shell.status == 0:
            return shell.stdout
        else:
            logging.error(shell.error())
            return ''

    def _fetch_lvs(self):
        shell = Shell('lvs', sudo=True)
        if shell.status == 0:
            return shell.stdout
        else:
            logging.error(shell.error())
            return ''

    def _fetch_vgs(self):
        shell = Shell('vgs -o vg_name,vg_size,vg_free', sudo=True)
        if shell.status == 0:
            return shell.stdout
        else:
            logging.error(shell.error())
            return ''

    def _fetch_mdstat(self):
        if os.path.isfile('/proc/mdstat'):
            try:
                content = open('/proc/mdstat', 'r').read()
                return content
            except:
                logging.error('Can\'t read /proc/mdstat')
                return ''
        return 'N/A'

    def _fetch_systemd(self):
        result = {'_main': {}, 'logind': {}}
        if os.path.isfile('/etc/systemd/logind.conf'):
            try:
                with open('/etc/systemd/logind.conf', 'r') as f:
                    for line in f:
                        data = line.split('=')
                        if not len(data) == 2:
                            continue
                        k, v = data[0].replace('#', ''), data[1]
                        result['logind'][k] = v
                if 'RemoveIPC' in result['logind']:
                    result['_main']['RemoveIPC'] = \
                        result['logind']['RemoveIPC']
            except:
                pass
        return result



# -*- coding: utf-8 -*-

import time
import logging
import signal
import sys

from mamonsu.first_look.start import run_first_look
from mamonsu.auto_tune.start import run_auto_tune
import mamonsu.lib.platform as platform
from mamonsu.lib.plugin import Plugin
from mamonsu.lib.config import Config
from mamonsu.lib.sender import *
from mamonsu.lib.senders import *
from mamonsu.plugins import *


class Supervisor(object):

    Running = True

    def __init__(self, config):
        self.Plugins = []
        self.config = config
        self._sender = Sender()
        self._senders = []

    def start(self):
        self._load_plugins()
        self._find_senders()
        self._set_senders()
        self._loop()

    def _load_plugins(self):
        for klass in Plugin.__subclasses__():
            plugin = klass(self.config)
            self.Plugins.append(plugin)

    def _find_senders(self):
        for plugin in self.Plugins:
            if plugin.is_sender():
                self._senders.append(plugin)
        if len(self._senders) == 0:
            raise RuntimeError('Can\'t find any senders')

    def _set_senders(self):
        for plugin in self.Plugins:
            plugin.update_sender(self._sender)
        self._sender.update_senders(self._senders)

    def _loop(self):
        plugin_errors, plugin_probes, last_error = 0, 0, ''
        while self.Running:
            for plugin in self.Plugins:
                if not plugin.is_alive():
                    plugin.start()
                    last_error = plugin.last_error_text
                    plugin_errors += 1
            time.sleep(10)
            # error counts
            plugin_probes += 1
            if plugin_probes >= 60:
                if plugin_errors > 0:
                    self._sender.send(
                        'mamonsu.plugin.errors[]',
                        'Errors in the last 60 seconds: {0}.\
                        Last error: {1}'.format(
                            plugin_errors, last_error))
                else:
                    self._sender.send('mamonsu.plugin.errors[]', '')
                plugin_errors, plugin_probes = 0, 0


def start():

    def quit_handler(_signo=None, _stack_frame=None):
        logging.info("Bye bye!")
        sys.exit(0)

    signal.signal(signal.SIGTERM, quit_handler)
    if platform.LINUX:
        signal.signal(signal.SIGQUIT, quit_handler)

    for arg in sys.argv:
        if arg == '--run-first-look' or arg == '--run-look'\
                or arg == 'first' or arg == '--first'\
                or arg == 'first-look' or arg == 'report':
            sys.argv.remove(arg)
            run_first_look()
            return

    for arg in sys.argv:
        if arg == 'auto-tune' or arg == '--auto-tune'\
                or arg == 'tune' or arg == 'autotune'\
                or arg == '--autotune':
            sys.argv.remove(arg)
            run_auto_tune()
            return

    config = Config()
    supervisor = Supervisor(config)

    try:
        logging.info("Start agent")
        supervisor.start()
    except KeyboardInterrupt:
        quit_handler()



from mamonsu.lib.plugin import Plugin


class Health(Plugin):

    def __init__(self, config):
        super(Health, self).__init__(config)
        self.counter = 0

    def run(self, zbx):
        zbx.send('mamonsu.plugin.keepalive[]', 0)
        self.counter += 1
        if self.counter > 9:
            self.log.info('=== Keep alive ===')
            self.counter = 0

    def items(self, template):
        # see supervisor.py:
        return template.item({
            'name': 'Mamonsu: plugin errors',
            'key': 'mamonsu.plugin.errors[]',
            'value_type': Plugin.VALUE_TYPE.text  # text
        }) + template.item({
            'name': 'Mamonsu: plugin keep alive',
            'key': 'mamonsu.plugin.keepalive[]'
        })

    def triggers(self, template):
        return template.trigger({
            'name': 'Mamonsu plugin errors '
            'on {HOSTNAME}. {ITEM.LASTVALUE}',
            'expression': '{#TEMPLATE:mamonsu.plugin.errors[].strlen()'
            '}&gt;1'
        }) + template.trigger({
            'name': 'Mamonsu nodata from {HOSTNAME}',
            'expression': '{#TEMPLATE:mamonsu.plugin.keepalive[]'
            '.nodata(180)}=1'
        })



from mamonsu.lib.plugin import Plugin


class Memory(Plugin):

    # colors
    # 1. physical memory
    # 2. virtual memory

    Items = [
        # zbx_key, meminfo_key, name, color
        ('apps', None,
            'Apps: User-space applications', 'CC0000'),

        ('buffers', 'Buffers',
            'Buffers: Block device cache and dirty', '00CC00'),

        ('swap', None,
            'Swap: Swap space used', '0000CC'),

        ('cached', 'Cached',
            'Cached: Parked file data (file content) cache', 'CC00CC'),

        ('unused', 'MemFree',
            'Free: Wasted memory', '000000'),

        ('slab', 'Slab',
            'Slab: Kernel used memory (inode cache)', 'CCCC00'),

        ('swap_cache', 'SwapCached',
            'SwapCached: Fetched unmod yet swap pages', '777777'),

        ('page_tables', 'PageTables',
            'PageTables: Map bt virtual and physical', '770000'),

        ('vmalloc_used', 'VmallocUsed',
            'VMallocUsed: vmaloc() allocated by kernel', '000077'),

        ('committed', 'Committed_AS',
            'Committed_AS: Total committed memory', '007700'),

        ('mapped', 'Mapped',
            'Mapped: All mmap()ed pages', 'DF0000'),

        ('active', 'Active',
            'Active: Memory recently used', '00DF00'),

        ('inactive', 'Inactive',
            'Inactive: Memory not currently used', '0000DF')
    ]

    def run(self, zbx):

        meminfo, result = {}, {}

        with open('/proc/meminfo', 'r') as f:
            for line in f:
                data = line.split()
                key, val = data[0], data[1]
                key = key.split(':')[0]
                meminfo[key] = float(val) * 1024

        for item in self.Items:
            zbx_key, meminfo_key = item[0], item[1]
            if meminfo_key is not None:
                result[zbx_key] = meminfo.get(meminfo_key) or 0
        result['apps'] = meminfo['MemTotal'] - result['unused'] \
            - result['buffers'] - result['cached'] - result['slab'] \
            - result['page_tables'] - result['swap_cache']
        result['swap'] = (meminfo.get('SwapTotal') or 0) \
            - (meminfo.get('SwapFree') or 0)

        for key in result:
            zbx.send('system.memory[{0}]'.format(key), result[key])

        del result, meminfo

    def items(self, template):
        result = ''
        for item in self.Items:
            result += template.item({
                'name': '{0}'.format(item[2]),
                'key': 'system.memory[{0}]'.format(item[0]),
                'units': 'b'
            })
        return result

    def graphs(self, template):
        items = []
        for item in self.Items:
            items.append({
                'key': 'system.memory[{0}]'.format(item[0]),
                'color': item[3]
            })
        graph = {
            'name': 'Memory overview', 'height': 400,
            'type': 1, 'items': items}
        return template.graph(graph)



# -*- coding: utf-8 -*-

from mamonsu.lib.plugin import Plugin
from .pool import Pooler
import time


class PgHealth(Plugin):

    def __init__(self, config):
        super(PgHealth, self).__init__(config)
        # сообщаем что у сервиса низкий аптайм, пока аптайм меньше 10 минут
        self.TriggerUptimeLessThen = self.config.fetch(
            'health', 'uptime', int)
        # алертим, если cache hit меньше чем %
        self.TriggerCacheHitLessThen = self.config.fetch(
            'health', 'cache', int)
        # счетчик, для сообщения в лог
        self.counter = 0

    def run(self, zbx):

        start_time = time.time()
        Pooler.query('select 1')
        zbx.send('pgsql.ping[]', (time.time() - start_time)*100)

        result = Pooler.query("select \
            date_part('epoch', now() - pg_postmaster_start_time())")
        zbx.send('pgsql.uptime[]', int(result[0][0]))

        result = Pooler.query('select \
            round(sum(blks_hit)*100/sum(blks_hit+blks_read), 2) \
            from pg_catalog.pg_stat_database')
        zbx.send('pgsql.cache[hit]', int(result[0][0]))

    def items(self, template):
        result = template.item({
            'name': 'PostgreSQL: ping',
            'key': 'pgsql.ping[]',
            'value_type': Plugin.VALUE_TYPE.numeric_float,
            'units': Plugin.UNITS.ms
        }) + template.item({
            'name': 'PostgreSQL: service uptime',
            'key': 'pgsql.uptime[]',
            'value_type': Plugin.VALUE_TYPE.numeric_unsigned,
            'units': Plugin.UNITS.uptime
        }) + template.item({
            'name': 'PostgreSQL: cache hit ratio',
            'key': 'pgsql.cache[hit]',
            'value_type': Plugin.VALUE_TYPE.numeric_unsigned,
            'units': Plugin.UNITS.percent
        })
        return result

    def graphs(self, template):
        items = [
            {'key': 'pgsql.cache[hit]'},
            {'key': 'pgsql.uptime[]', 'color': 'DF0101', 'yaxisside': 1}
        ]
        graph = {'name': 'PostgreSQL uptime', 'items': items}
        return template.graph(graph)

    def triggers(self, template):
        result = template.trigger({
            'name': 'PostgreSQL service was restarted on '
            '{HOSTNAME} (uptime={ITEM.LASTVALUE})',
            'expression': '{#TEMPLATE:pgsql.uptime[].last'
            '()}&lt;' + str(self.TriggerUptimeLessThen)
        }) + template.trigger({
            'name': 'PostgreSQL cache hit ratio too low on '
            '{HOSTNAME} ({ITEM.LASTVALUE})',
            'expression': '{#TEMPLATE:pgsql.cache[hit].last'
            '()}&lt;' + str(self.TriggerCacheHitLessThen)
        })
        return result



# -*- coding: utf-8 -*-

import os
import sys
import logging

import mamonsu.lib.platform as platform
from mamonsu.plugins.pgsql.pg8000.core import ProgrammingError
from mamonsu.plugins.pgsql.pool import Pooler


class AutoTune(object):

    def __init__(self, args):

        if not self._is_connection_work():
            logging.error('Can\'t connect to PostgreSQL')
            sys.exit(5)

        self.args = args
        self._memory()
        self._auto_vacuum()
        self._bgwriter()
        self._configure_pgbadger()
        logging.info('Auto-tune PostgreSQL config: completed')
        self._reload_config()

    def _memory(self):
        if platform.WINDOWS:
            logging.info("No memory config for windows")
            return
        sysmemory = self._get_total_mem()
        if sysmemory == 0:
            return

        self._run_query(
            "alter system set shared_buffers to '{0}'".format(
                self._humansize_and_round_bytes(sysmemory/4)))
        self._run_query(
            "alter system set effective_cache_size to '{0}'".format(
                self._humansize_and_round_bytes(3*sysmemory/4)))
        self._run_query(
            "alter system set work_mem to '{0}'".format(
                self._humansize_and_round_bytes(sysmemory/30)))
        self._run_query(
            "alter system set maintenance_work_mem to '{0}'".format(
                self._humansize_and_round_bytes(sysmemory/5)))

    def _auto_vacuum(self):
        self._run_query(
            "alter system set autovacuum_max_workers to 20")
        self._run_query(
            "alter system set autovacuum_analyze_scale_factor to 0.01")
        self._run_query(
            "alter system set autovacuum_vacuum_scale_factor to 0.02")
        self._run_query(
            "alter system set vacuum_cost_delay to 1")

    def _bgwriter(self):
        self._run_query(
            "alter system set bgwriter_delay to 10")
        self._run_query(
            "alter system set bgwriter_lru_maxpages to 800")

    def _configure_pgbadger(self):
        if not self.args.pgbadger:
            return
        self._run_query(
            "alter system set logging_collector to on")
        self._run_query(
            "alter system set log_filename to 'postgresql-%%a.log'")
        self._run_query(
            "alter system set log_checkpoints to on")
        self._run_query(
            "alter system set log_connections to on")
        self._run_query(
            "alter system set log_disconnections to on")
        self._run_query(
            "alter system set log_lock_waits to on")
        self._run_query(
            "alter system set log_temp_files to 0")
        self._run_query(
            "alter system set log_autovacuum_min_duration to 0")
        self._run_query(
            "alter system set log_line_prefix to \
                '%%t [%%p]: [%%l-1] db=%%d,user=%%u,app=%%a,client=%%h '")

    def _reload_config(self):
        if not self.args.reload_config:
            return
        logging.info('Reload config...')
        self._run_query('select pg_catalog.pg_reload_conf()')

    def _is_connection_work(self):
        try:
            Pooler.query('select 1')
            return True
        except Exception as e:
            logging.error('Test query error: {0}'.format(e))
            return False

    def _get_total_mem(self):
        if os.path.isfile('/proc/meminfo'):
            try:
                with open('/proc/meminfo', 'r') as f:
                    for line in f:
                        data = line.split()
                        if not len(data) == 3:
                            continue
                        if data[0] == 'MemTotal:' and data[2] == 'kB':
                            return int(data[1]) * 1024
            except:
                logging.error('Can\'t read /proc/meminfo')
        return 0

    def _run_query(self, query='', exit_on_fail=True):
        try:
            Pooler.query(query)
        except ProgrammingError as e:
            if '{0}'.format(e) == 'no result set':
                return
        except Exception as e:
            logging.error('Query {0} error: {1}'.format(query, e))
            if exit_on_fail:
                sys.exit(6)

    def _humansize_and_round_bytes(self, nbytes):
        suffixes = ['', 'kB', 'MB', 'GB', 'TB']
        if nbytes < 1024:
            return str(nbytes)
        i = 0
        while nbytes >= 1024 and i < len(suffixes)-1:
            nbytes /= 1024.
            i += 1
        f = ('%.2f' % nbytes).rstrip('0').rstrip('.')
        return '%s%s' % (int(round(float(f))), suffixes[i])
